﻿#Version 2.7

Function Add-Type-If-New
{
    [CmdletBinding()]
    Param($classname, $code)
    if(-not ($classname -as [type]))
    {
        add-type -TypeDefinition $code
        return
    }

    Write-Debug "Found $($classname), skipping adding type"
}

Add-Type-If-New 'Velostrata.UserScript.Platform' @"
   namespace Velostrata.UserScript
   {
       [System.Flags]
       public enum Platform : uint
       {
	        //Unknown = 0x0, // Commented are not user facing
	        Aws = 0x01,
	        Azure = 0x02,
	        Gcp = 0x010,
	        All = 0xFFFFFFFF
       }
    }
"@

Add-Type-If-New 'Velostrata.UserScript.TaskType' @"
   namespace Velostrata.UserScript
   {       
       public enum TaskType : uint
       {
            RunOnce = 1,
            RunAlways = 2,
            //DoNotExecute =3, 
       }
   }
"@
Add-Type-If-New 'Velostrata.UserScript.MachineState' @"
   namespace Velostrata.UserScript
   {
       public enum MachineState : uint
       {
            //Unknown = 0,
            Origin = 1,
            Velostrata = 2,
            Detach = 3,
       }
   }
"@

$ScriptsPath = 'c:\Program Files\Velostrata\UserScripts'

$VelostrataServicePath = 'HKLM:\SYSTEM\CurrentControlSet\Services\VelostrataService' 
$UserTasksPath = Join-Path $VelostrataServicePath 'UserTasks' 
$OriginPath = Join-Path $UserTasksPath "Origin"
$DetachPath = Join-Path $UserTasksPath "Detach"
$VelostrataPath = Join-Path $UserTasksPath "Velostrata"


<# 
 .Synopsis
  Install a user script task on a Velostrata machine.

 .Description
  Adds a script that can be executed on specified states, such as when the machine is on-premises, 
  detached or moved to cloud in Velostrata. A specific Cloud ("Platform") can be specified such as Azure or AWS. 
  Also the user can control whether the task is executed once or always. Only powershells are supported (.ps1).
  The tasks will be executed on machine boot.

 .Parameter FilePath
  Path to a PowerShell script file that will be executed by the user task. 
  Please note, The script is copied to the Velostrata user scripts 
  directory. If the script is dependent on other files, notice the paths are non-relative, or copy the dependant files to 
  the user scripts directory.

 .Parameter TaskName
  The task name, must be unique and will also show in the logs when executed.
  The tasks are executed in ASCII alphabetic order, "2020-Prepare-Boot" will be executed before "Prepare-Active-Directory"

 .Parameter MachineState
  The machine's state in the Velostrata machine lifecycle. This allows to specify whether the script will be executed when
  the machine is On-Premises (Origin), Detached or Moved to Cloud (Velostrata)

 .Parameter Platform
  The script will only be executed on machines running on the following platform. If MachineState is on-premises, only 
  "All" is supported.
  This parameter is optional, if it is not supplied, it is set to All by default.

 .Parameter TaskType
  Whether the task will be executed once or multiple times

 .Parameter Optional
  Whether the task will be skipped if shutdown is in progress

 .Parameter Overwrite
  Whether existing tasks will be overwritten  

 .Example
   
   Install-VelosUserScript -FilePath .\testscript.ps1 -TaskName Debugging-Task -MachineState Origin -TaskType RunOnce -Platform All

   # Runs testscript.ps1 on machines that are on-premises. Run the task only on the first reboot.

 .Example
      
   Install-VelosUserScript -FilePath .\resetWMI.ps1 -TaskName ResetWMI -MachineState Detach -TaskType RunOnce -Platform Aws -Overwrite

   # Run resetWmi.ps1 on the first boot in detached machines that are in Amazon AWS cloud. 
   # Existing tasks will be overwritten.

 .Example
   
   Install-VelosUserScript -FilePath c:\scripts\Fetch-Eventlogs.ps1 -TaskName FetchEventLogs -MachineState Velostrata -TaskType RunAlways -Platform All -Optional

   # Run Fetch-Eventlogs.ps1 on machines that were moved to cloud by Velostrata, on every reboot, on all platforms. 
   # The task will not be executed if the system is in the middle of a shutdown.
#>
Function Install-VelosUserScript
{
[CmdletBinding()]
Param(
    [Parameter(Mandatory=$true)]
    [string]$FilePath,

    [Parameter(Mandatory=$true)]
    [string]$TaskName,

    [Parameter(Mandatory=$true)]
    [Velostrata.UserScript.MachineState]$MachineState,

    [Parameter(Mandatory=$true)]
    [Velostrata.UserScript.TaskType]$TaskType,

    [Parameter(Mandatory=$false)]
    [Velostrata.UserScript.Platform]$Platform = [Velostrata.UserScript.Platform]::All,

    [Parameter(Mandatory=$false)]
    [switch]$Optional,

    [Parameter(Mandatory=$false)]
    [switch]$Overwrite
)

    ValidateInputInstall $FilePath $TaskName $MachineState $Platform $TaskType $($Overwrite.IsPresent)
    SetupPaths
    SetupKeys
    CopyScript $FilePath 
    AddRegistry $FilePath $TaskName $MachineState $Platform $TaskType $($Optional.IsPresent)

}

Function ValidateInputInstall($FilePath, $TaskName, $MachineState, $Platform, $TaskType, $Overwrite)
{
    $TaskPath = GetTaskPath $TaskName $MachineState
    
    if((-not $Overwrite) -and (Test-Path $TaskPath)) 
    {
        throw "A task with the same name already exists"
    }
    
    if(-Not (Test-Path $FilePath))
    {
        throw "Could not find script file, cannot add task"
    }

    if([System.IO.Path]::GetExtension($FilePath) -ne ".ps1")
    {
        throw "File extension is not .ps1, only PowerShell scripts are supported"
    }

    if (($MachineState -eq [Velostrata.UserScript.MachineState]::Origin) `
         -and ($Platform -ne [Velostrata.UserScript.Platform]::All))
    {
        throw "Machine state Origin only supports the platform parameter set to All. [MachineState=$($MachineState)  Platform=$($Platform)]"
    }

    if (($MachineState -ne [Velostrata.UserScript.MachineState]::Velostrata) `
        -and ($TaskType -eq [Velostrata.UserScript.TaskType]::RunAlways))
    {
        throw "RunAlways Task types are only supported on Velostrata machine state"   
    }
}


Function SetupPaths()
{
    if(-Not (Test-Path $ScriptsPath))
    {
        New-Item -ItemType Directory -Path $ScriptsPath | Out-Null
        return
    }

    if (-not ((Get-Item $ScriptsPath) -is [System.IO.DirectoryInfo]))
    {
        throw "A file with the same name as the directory already exists [$($ScriptsPath)]"
    }
}

Function CreateKeyIfMissing($KeyPath)
{
    if(-Not (Test-Path $KeyPath))
    {
        Write-Debug "Creating key $($KeyPath)..."
        New-Item -Path $KeyPath -ItemType Directory | Out-Null
    }
}

Function SetupKeys()
{
    CreateKeyIfMissing $VelostrataServicePath
    CreateKeyIfMissing $UserTasksPath

    CreateKeyIfMissing $OriginPath
    CreateKeyIfMissing $DetachPath
    CreateKeyIfMissing $VelostrataPath
}

Function CopyScript($ScriptName)
{
    Write-Output "Copying script $($ScriptName) to $($ScriptsPath)"
    Copy-Item -Path $ScriptName -Destination $ScriptsPath | Out-Null
}

Function AddRegistry($FilePath, $TaskName, $MachineState, $Platform, $TaskType, $Optional)
{
    $TaskPath = GetTaskPath $TaskName $MachineState
    $FileName = [System.Io.Path]::GetFileName($FilePath)
   
    Write-Output "Adding Task in $($TaskPath) ..."

    Write-Output ""
    Write-Output "FileName = $($FileName)"
    Write-Output "TaskName = $($TaskName)"
    Write-Output "MachineState = $($MachineState)"
    Write-Output "Platform = $($Platform)"
    Write-Output "TaskType = $($TaskType)"
    Write-Output "Optional = $($Optional)"
    
    CreateKeyIfMissing $TaskPath

    New-ItemProperty -Path $TaskPath -Name "Execute" -PropertyType "String" -Value $FileName -Force | Out-Null
    New-ItemProperty -Path $TaskPath -Name "Type" -PropertyType "Dword" -Value ([uint32] $TaskType) -Force | Out-Null
    New-ItemProperty -Path $TaskPath -Name "Platform" -PropertyType "DWord" -Value ([uint32] $Platform) -Force | Out-Null
    New-ItemProperty -Path $TaskPath -Name "Optional" -PropertyType "DWord" -Value ([uint32] $Optional) -Force | Out-Null

    Write-Output "Done"
}

Function GetTaskPath($TaskName, $MachineState)
{
    $MachineStatePath = GetKeyPathByMachineState $MachineState
    $TaskPath = Join-Path $MachineStatePath $TaskName
    return $TaskPath
}

Function GetKeyPathByMachineState($MachineState)
{   
    $result = Join-Path $UserTasksPath $($MachineState.ToString())
    return $result
}


<# 
 .Synopsis
  Removes a user script task on a Velostrata machine.

 .Description
  Remove a user script under the context of a specific machine state from the local machine.

 .Parameter TaskName
  The task name, must be unique and will also show in the logs when executed.
  The tasks are executed in ASCII alphabetic order, "2020-Preparation-Script" will be executed before "Prepare-Active-Directory"

 .Parameter MachineState
  The machine's state in the Velostrata machine lifecycle. This allows to specify whether the script will be executed when
  the machine is On-Premises (Origin), Detached or Moved to Cloud (Velostrata)

 .Parameter DeleteScriptFile
  If this switch is specified, the task's script will be deleted as well from the Velostrata user scripts directory

 .Example
   
   Remove-VelosUserScript -TaskName Debugging-Task -MachineState Origin

   # Removes the task named Debugging-Task that runs on Origin

 .Example
      
   Remove-VelosUserScript -TaskName ResetWMI -MachineState Detach -DeleteScriptFile

   # Removes the task named ResetWMI running on Detach, and also delete its accompying script file from the Velostrata user scripts directory

#>
Function Remove-VelosUserScript
{
[CmdletBinding()]
Param(
    [Parameter(Mandatory=$true)]
    [string]$TaskName,
    
    [Parameter(Mandatory=$true)]
    [Velostrata.UserScript.MachineState]$MachineState,
    
    [Parameter(Mandatory=$false)]
    [Switch]$DeleteScriptFile
)

    ValidateInputDelete $TaskName $MachineState 
    $TaskScriptName = GetScriptNameFromTask $TaskName $MachineState
    
    DeleteKey
    
    if($DeleteScriptFile.IsPresent)
    {
        DeleteScript $TaskScriptName
    } 
  
    Write-Output("Done")
}

Function DeleteScript($TaskScriptName)
{
    $FullScriptPath = Join-Path $ScriptsPath $TaskScriptName
    Write-Output "Deleting script $($FullScriptPath)"
    Remove-Item -Path $FullScriptPath
}

Function DeleteKey
{
   $TaskPath = GetTaskPath $TaskName $MachineState
   Write-Output "Deleting key $($TaskPath)"
   Remove-Item -Path $TaskPath
}

Function GetScriptNameFromTask($TaskName, $MachineState)
{
    $TaskPath = GetTaskPath $TaskName $MachineState
    $TaskScriptName = (Get-ItemProperty -Path $TaskPath).Execute
    return $TaskScriptName
}

Function ValidateInputDelete($TaskName, $MachineState)
{
    $TaskPath = GetTaskPath $TaskName $MachineState
    if(-Not (Test-Path $TaskPath))
    {
        throw "Task does not exist"
    }
}

Export-ModuleMember -Function Install-VelosUserScript
Export-ModuleMember -Function Remove-VelosUserScript
